package decrypto;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.awt.*;
import javax.swing.*;

import decrypto.util.*;
import decrypto.dictattack.*;
import decrypto.genetic.*;

public class DecryptoCGI
{
    SolutionSet        set;
    GetOpt             opt = new GetOpt();
    Puzzle             puzzle;
    Solver             solver;
    Random             r = new Random();
    Dict               dict;
    double             bestscore = -Double.MAX_VALUE;
    
    class FailsafeThread extends Thread
    {
	int seconds;

	public FailsafeThread(int seconds)
	{
	    this.seconds = seconds;
	    setDaemon(true);
	}

	public void run()
	{
	    try {
		Thread.sleep(seconds*1000);

		System.out.println("An error has occured. Logging it!");

		BufferedWriter outs = new BufferedWriter(new FileWriter("/var/tmp/decrypto_failsafe.log", true));
		outs.write("***********************\n");
		outs.write("DATE:         "+new Date()+"\n");
		outs.write("DICTIONARY:   "+opt.getString("dictionary")+"\n");
		outs.write("TRUST SPACES: "+opt.getBoolean("trustspaces")+"\n");
		outs.write("PUZZLE:       "+puzzle.cipherText+"\n");
		outs.write("CLUES:        "+puzzle.clues+"\n\n");


		if (solver != null && solver.getException()!=null) {
		    Exception e = solver.getException();
		    e.printStackTrace(new PrintWriter(outs));
		}
		
		outs.close();

		System.exit(-1);
	    } catch (InterruptedException ex) {
	    } catch (IOException ex) {
	    }
	}
    }

    String escape(String s)
    {
	s=s.replace("\"","'");
	s=s.replace("\n"," ");
	s=s.replace("\r","").trim();
	s=s.replace("  "," &nbsp;");

	return s;
    }

    public class MyDecryptoListener implements SolverListener
    {
	boolean finished = false;

	void msg(String s)
	{
	    System.out.println("<script>msg(\""+s+"\");</script>");
	}

	public void solverMessage(String s)
	{
	    msg(s);
	}

	public void solverFinished(double elapsedTime)
	{
	    String msg = "";

	    msg = String.format("Finished (%.3f seconds)", elapsedTime);

	    solverMessage(msg);

	    synchronized(this) {
		finished = true;
		this.notifyAll();
	    }
	}

	public void solverSolution(MapSet mapset, int spaces[], boolean fullSolution, double score)
	{
	    if (set.wouldAdd(score))
		{
		    String ss = puzzle.lang.translateString(puzzle.rawCipherText, mapset, spaces);
		    Solution sol = new Solution(ss, mapset, score);

		    boolean best = set.size() > 0 ? set.getElement(0).score < score : false;		    
		    boolean added = set.add(sol);

		    if (added)
			{
			    if (score > bestscore) {
				bestscore = score;
				System.out.println(String.format("<script>sol(%.4f,\"%s\");</script>",
								 sol.score, escape(sol.solution)));
			    }
			}
		}
	}

	String lastprg = "";
	int nprogs = 0;
	public void solverProgress(double progress)
	{
	    String newprg = String.format("<script>prg(%.2f);</script>", progress);
	    if (!newprg.equals(lastprg))
		{
		    nprogs ++;
		    if (nprogs < 100 || (nprogs&(nprogs-1))==0)
			System.out.println(newprg);
		    lastprg=newprg;
		}
	}
    }

    public static void main(String args[])
    {
	DecryptoCGI app = new DecryptoCGI();
	app.run(args);
    }

    public DecryptoCGI()
    {
	opt.addBoolean('h', "help", false, "Show help");
	opt.addString('d', "dictionary", "english-standard.dat", "Dictionary to use");
	opt.addString('c', "clues", "", "Clues, e.g.: A=C, M=R, JXJ=DAD");
	opt.addDouble('\0', "timelimit", 10, "Maximum CPU time");
	opt.addInt('\0', "setsize", 100, "Maximum solutions to store");
	opt.addString('\0',"puzzlefile", "", "File containing the puzzle");
	opt.addBoolean('\0',"scramble", false, "Scramble the puzzle and exit");
	opt.addBoolean('\0',"trustspaces", true, "Trust spaces");
    }

    public void run(String args[])
    {
	opt.parse(args);

	if (opt.getBoolean("help"))
	    {
		opt.doHelp();
		return;
	    }

	try {
	    dict = new Dict(opt.getString("dictionary"));
	} catch (IOException ex) {
	    System.out.println("Couldn't open dictionary: "+ex);
	    return;
	}

	set = new SolutionSet(opt.getInt("setsize"));

	String ciphertext;
	String clues;

	if (opt.wasSpecified("puzzlefile"))
	    {
		StringBuffer ct = new StringBuffer();
		StringBuffer cs = new StringBuffer();

		try {
		    BufferedReader ins = new BufferedReader(new FileReader(opt.getString("puzzlefile")));
		    String line;
		    
		    while ((line = ins.readLine())!=null)
			{
			    if (line.contains("="))
				cs.append(line+" ");
			    else
				ct.append(line+" ");
			}
		} catch (IOException ex) {
		    System.out.println("Couldn't read puzzle file: "+ex);
		}

		ciphertext = ct.toString();
		clues = cs.toString();
	    }
	else
	    {
		ArrayList<String> extraArgs = opt.getExtraArgs();
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < extraArgs.size(); i++)
		    sb.append(extraArgs.get(i)+" ");

		ciphertext = sb.toString();
		clues = opt.getString("clues");
	    }

	if (opt.getBoolean("scramble"))
	    {
		puzzle = new Puzzle(dict, ciphertext, clues, true, true);
		puzzle.scramble(false);
		System.out.println(puzzle.cipherText);
		System.out.println(puzzle.clues);
		return;
	    }

	new FailsafeThread((int) opt.getDouble("timelimit")+5).start();

	if (opt.getBoolean("trustspaces")) {
	    puzzle = new Puzzle(dict, ciphertext, clues, true, true);
	    solver = new DictionaryAttackSolver(puzzle);

	} else {
	    puzzle = new Puzzle(dict, ciphertext, clues, false, true);
	    solver = new GeneticSolver(new GAProblem(puzzle));
	}

	MyDecryptoListener mdl = new MyDecryptoListener();
	solver.addListener(mdl);

	solver.start(opt.getDouble("timelimit"));

	synchronized(mdl) {
	    while (!mdl.finished) {
		try {
		    mdl.wait();
		} catch (InterruptedException ex) {
		}
	    }
	}

	for (int i = 0; i < set.size(); i++)
	    {
		Solution sol = set.getElement(i);

		System.out.println(String.format("<script>solsum(%d,%.3f,\"%s\");</script>",
						 i, sol.score, escape(sol.solution)));
	    }

	double elapsedTime = 0;

	System.out.println("<script>elapse("+elapsedTime+")</script>");
    }
}
