package decrypto;

import java.util.*;

public class Finisher
{
    Puzzle   puzzle;
    int      startmap[];     // The best (partial) solution from which we begin our work.
    int      ourmap[];       // Our state.

    Language lang;

    byte     cletters[];
    byte     pletters[];

    Random rand = new Random();

    public Finisher(Puzzle puzzle, MapSet _map)
    {
	this.puzzle = puzzle;
	this.lang = puzzle.lang;

	// ciphertext letters that do not have a plaintext mapping
	byte _cletters[] = new byte[lang.letterCount()];
	int ncletters = 0;
	
	// plaintext letters that do not have a ciphertext mapping
	byte _pletters[] = new byte[lang.letterCount()];
	int npletters = 0;
	
	boolean mapped[] = new boolean[lang.letterCount()];
	
	// look at each ciphertext letter and its current mapping.
	for (int i = 0; i < lang.letterCount(); i++) {
	    
	    int mi = _map.getMapping((byte) i);
	    
	    if (mi >= 0)
		mapped[_map.getMapping((byte) i)] = true;
	    else if (puzzle.histogram[i] > 0)
		_cletters[ncletters++] = (byte) i;
	}
	
	for (int i = 0; i < lang.letterCount(); i++) {
	    if (!mapped[i])
		_pletters[npletters++] = (byte) i;
	}

	cletters = new byte[ncletters];
	for (int i = 0; i < ncletters; i++)
	    cletters[i] = _cletters[i];

	pletters = new byte[npletters];
	for (int i = 0; i < npletters; i++)
	    pletters[i] = _pletters[i];

	startmap = _map.makeMap();
    }

    public MapSet bruteForce()
    {
	boolean verbose = false;

	if (cletters.length ==0)
	    return new MapSet(startmap);

	ourmap = new int[startmap.length];
	for (int i = 0; i < startmap.length; i++)
	    ourmap[i] = startmap[i];

	int iters = (int) Math.pow(pletters.length, cletters.length);

	long starttime = System.currentTimeMillis();
	if (verbose) {
	    System.out.printf("===Greedy Finisher: cipherletters: %d plainletters: %d (%d iters)\n", 
			      cletters.length, pletters.length, iters);
	    System.out.printf("%15.5f : %s\n", lang.computeScore(puzzle.cipherBytes, ourmap)/puzzle.numCiphertextLetters,
			      lang.translateString(puzzle.cipherText, new MapSet(ourmap), null));
	}

	if (iters > 30000) {
	    return new MapSet(ourmap);
	}

	// newmap: for each unknown ciphertext letter, what plaintext
	// letter does it map to?
	int newmap[] = new int[cletters.length];
	int bestmap[] = new int[ourmap.length];

	double bestscore = -Double.MAX_VALUE;

	for (int iter = 0; iter < iters; iter++) {
	    int tmp = iter;
	    boolean good = true;

	    for (int i = 0; i < cletters.length; i++) {
		newmap[i] = tmp % pletters.length;
		tmp /= pletters.length;

		for (int j = 0; j < i; j++) {
		    if (newmap[j]==newmap[i])
			good = false;
		}

		if (!puzzle.initialMap.isMappingOkay(cletters[i], pletters[newmap[i]])) {
		    good = false;
		    break;
		}

		ourmap[cletters[i]] = pletters[newmap[i]];
	    }

	    if (!good)
		continue;

	    double score = lang.computeScore(puzzle.cipherBytes, ourmap);
	    if (score > bestscore) {
		bestscore = score;
		for (int i = 0; i < ourmap.length; i++)
		    bestmap[i] = ourmap[i];
	    }
	}

	if (bestmap == null)
	    return new MapSet(startmap);

	if (verbose) {
	    System.out.printf("%15.5f : %s\n", lang.computeScore(puzzle.cipherBytes, bestmap)/puzzle.numCiphertextLetters,
			      lang.translateString(puzzle.cipherText, new MapSet(bestmap), null));
	    
	    long endtime = System.currentTimeMillis();
	    System.out.printf("time: %15.5f\n", (endtime-starttime)/1000.0);
	}

	return new MapSet(bestmap);
    }

    /** Create a new random mapping between the unknown cletters and
	pletters. (Keep all other characters in the same place!
    **/
    public void randomize()
    {
	for (int i = 0; i < cletters.length; i++) {
	    
	    while (true) {
		int j = rand.nextInt(pletters.length);
		boolean okay = true;

		// don't allow a second mapping to the same plaintext
		// letter
		for (int k = 0; k < i; k++) {
		    if (ourmap[cletters[k]]==pletters[j])
			okay = false;
		}

		if (okay) {
		    ourmap[cletters[i]] = pletters[j];
		    break;
		}
	    }
	}
    }

    public void iterate()
    {
	
    }
}
