package decrypto;

import java.util.*;

/** represents a single cryptogram or patristocrat. **/
public class Puzzle
{
    public String rawCipherText; // the ciphertext as entered by the
				 // user. May contain bad formatting
				 // and/or punctuation.

    public String cipherText;    // ciphertext with multiple spaces
				 // replaced with single space. No
				 // punctuation. If !trustSpaces, no spaces.

    public int numCiphertextLetters; // how many non-whitespace in ciphertext?
    public int numUniqueCiphertextLetters; // how many *different* ciphertext letters are used?
    public byte[] cipherBytes;   // cipherBytes.length==cipherText.length,
				 // with each character (including
				 // spaces) mapped by
				 // lang.charToByte(). Note: letters
				 // are mapped [0,N-1], and spaces are
				 // N. If !trustspaces, then there are
				 // no spaces.
    
    public boolean trustSpaces;  // Do we trust the position of the
				 // spaces in the ciphertext? (I.e.,
				 // is this a patristocrat?)
    
    public String clues;         // human-readable clues. (e.g., "A=R RZN=THE")
    public MapSet initialMap;    // Map implied by human clues.

    public int[] histogram;      // for each ciphertext letter, how many times does it appear?
    public byte[] historder;     // ciphertext letters, in decreasing order of frequency

    public Dict dict;            // dictionary selected by user. 
    public Language lang;        // language specified by dictionary.

    public ArrayList<Word> words; // The cipher words parsed from this puzzle.

    //////////////////////////////////////////////////////////////////////////
    public Puzzle(Dict dict, String _ctext, String _clues, boolean trustSpaces, boolean allowIdentityWhenSolving)
    {
	this.rawCipherText = _ctext;
	this.trustSpaces = trustSpaces;
	this.dict = dict;
	this.lang = dict.getLanguage();
	this.clues = despaceString(_clues.trim());

	if (trustSpaces)
	    this.cipherText = onlyAlphaSpaceCharacters(despaceString(_ctext.trim()));
	else
	    this.cipherText = onlyAlphaCharacters(despaceString(_ctext.trim()));

	this.cipherBytes = lang.cipherTextToCipherBytes(this.cipherText);

	// set up the initialMap based on the user-specified clues
	this.initialMap = new MapSet(lang.letterCount());
	this.initialMap.setFullSet();
	if (!allowIdentityWhenSolving) {
	    for (int i = 0; i < lang.letterCount(); i++) 
		initialMap.forbidMapping((byte) i, (byte) i);
	}
	handleClues(this.clues);

	// how many alphabetic characters in ciphertext? We us this
	// for normalizing log probabilities.
	numCiphertextLetters = 0;
	for (int i = 0; i < cipherBytes.length; i++) {
	    if (cipherBytes[i] < lang.letterCount())
		numCiphertextLetters++;
	}

	numUniqueCiphertextLetters = 0;
	if (true) {
	    boolean present[] = new boolean[lang.letterCount()];
	    for (int i = 0; i < cipherBytes.length; i++) {
		if (cipherBytes[i]==lang.letterCount())
		    continue;
		if (!present[cipherBytes[i]])
		    numUniqueCiphertextLetters++;
		present[cipherBytes[i]] = true;
	    }
	}

	////////////////////////////////////////////////////
	// remove duplicate words
	if (true ) {
	    words = new ArrayList<Word>();
	    // we want words despaced, but keep extra punctuation in
	    // there for now because some punctuation will be used as
	    // word separators.  i.e., *don't* pass in either
	    // rawCipherText (extra spaces) OR cipherText (punctuation
	    // removed).
	    ArrayList<Word> rawwords = lang.cipherTextToWords(despaceString(rawCipherText));
	    Word.WordComparator wcomp = new Word.WordComparator();
	    
	    for (Word w : rawwords) {
		boolean existsAlready = false;
		
		for (Word uw : words)  {
		    if (wcomp.compare(w, uw) == 0) {
			existsAlready = true;
			break;
		    }
		}
		
		if (!existsAlready)
		    words.add(w);
	    }
	}

	////////////////////////////////////////////////////
	// Look up each word in the dictionary for a list of
	// candidates
	for (Word w : words) {
	    w.candidates = Candidate.toCandidates(dict.lookup(w.pattern), lang);
	    if (w.candidates == null || w.candidates.length == 0)
		w.enabled = false; 
	}

	////////////////////////////////////////////////////
	// Compute histogram of ciphertext letters
	ArrayList<Hist> hists = new ArrayList<Hist>();
	for (int i = 0; i < lang.letterCount(); i++)
	    hists.add(new Hist(i));

	for (Word w : words) {
	    for (int i = 0; i < w.word.length; i++)
		hists.get(w.word[i]).count++;
	}
	
	histogram = new int[lang.letterCount()];
	
	for (int i = 0; i < lang.letterCount(); i++)
	    histogram[i] = hists.get(i).count;
	
	Collections.sort(hists);

	historder = new byte[lang.letterCount()];

	for (int i = 0; i < lang.letterCount(); i++) {
	    historder[i] = (byte) hists.get(i).letter;
	}

	////////////////////////////////////////////////////
	if (false) {
	    System.out.printf("%15s  %15s  %10s  %s\n", "ciphertext", "pattern", "cands", "enabled");
	    for (Word w: words) {
		System.out.printf("%15s  %15s  %10d  %b\n", w.text, lang.bytesToString(w.pattern), w.candidates.length, w.enabled);
	    }
	}
    }

    static class Hist implements Comparable<Hist>
    {
	int count;
	int letter;
	public Hist(int letter)
	    {
		this.letter = letter;
	    }
	
	public int compareTo(Hist h)
	{
	    return h.count - count;
	}
    }
    
    /** If they entered the puzzle in with a lot of spaces, T H E  E M E R G E N...,
	remove the extra spaces. **/
    public static String despaceString(String s)
    {
	// count the spaces. 
	int spacecount = 0;

	for (int i = 0; i < s.length(); i++) {
	    if (s.charAt(i)==' ')
		spacecount++;
	}

	double frac = ((double) spacecount)/s.length();

	// few enough characters are spaces. I think the puzzle is encoded properly
	// (threshold = 0.4 so that "A = B" will trigger the despacing logic.
	if (frac < .4)
	    return s;

	// now count the number of non-space characters which are followed by a space.
	if (s.length() > 8) {
	    spacecount = 0;
	    for (int i = 0; i < s.length()-2; i++) {
		if (s.charAt(i)==' ' && s.charAt(i+2)==' ')
		    spacecount++;
	    }
	    
	    frac = ((double) spacecount)/(s.length()-2);
	    if (frac < .4)
		return s;
	}

	// remove any occurence of one space, replace any occurence of
	// more than one space with just one space.
	StringBuffer sb = new StringBuffer();
	int spacerun = 0;
	for (int i = 0; i < s.length(); i++) {
	    if (s.charAt(i)==' ')
		{
		    spacerun++;
		    continue;
		}
	    
	    // it's not a space.
	    
	    // if we didn't see many spaces, just encode this directly out.
	    if (spacerun<=1) {
		sb.append(s.charAt(i));
	    } else {
		sb.append(' ');
		sb.append(s.charAt(i));
	    }
	    
	    spacerun = 0;
	    
	}
	return sb.toString();
    }


    protected Puzzle()
    {
    }

    public Puzzle copy()
    {
	Puzzle dp = new Puzzle();
	dp.rawCipherText = rawCipherText;
	dp.cipherText = cipherText;
	dp.cipherBytes = cipherBytes;
	dp.trustSpaces = trustSpaces;
	dp.clues = clues;
	dp.initialMap = initialMap.copy();
	dp.dict = dict;
	dp.lang = lang;

	dp.historder = historder;
	dp.histogram = histogram;

	dp.words = new ArrayList<Word>();
	for (Word w : words)
	    dp.words.add(w.copy());

	return dp;
    }

    void handleClues(String clues)
    {
	this.clues = clues;
	String toks[]=clues.split("[\\s,]+");

	for (int i = 0; i < toks.length; i++) {
	    if (toks[i].length()==0)
		continue;

	    if (toks[i].contains("!=")) {
		// negative assertion: the left hand side does NOT
		// equal the right hand side.
		String kv[] = toks[i].split("!=");
		if (kv==null || kv.length!=2) {
		    System.out.println("Unable to parse clue: "+toks[i]);
		    continue;
		}

		// A != ABC
		// A does not match any of A, B, or C
		if (kv[0].length()==1) {
		    for (int j = 0; j < kv[1].length(); j++)
			initialMap.forbidMapping(lang.charToByte(kv[0].charAt(0)),
						 lang.charToByte(kv[1].charAt(j)));
		    continue;
		}

		// ABC != DEF
		// A !=D, B!=E, C!=F
		if (kv[0].length() != kv[1].length()) {
		    System.out.println("Invalid clue: "+toks[i]);
		    continue;
		}

		for (int j = 0; j < kv[0].length() ; j++) {
		    byte k = lang.charToByte(kv[0].charAt(j));
		    byte v = lang.charToByte(kv[1].charAt(j));
		    initialMap.forbidMapping(k, v);
		}
		
	    } else {
		// Ordinary (positive) mapping. (With equality)
		String kv[] = toks[i].split("=");
		if (kv==null || kv.length!=2) {
		    System.out.println("Unable to parse clue: "+toks[i]);
		    continue;
		}
		
		if (kv[0].length() != kv[1].length()) {
		    System.out.println("Must have same number of letters before and after equal sign: "+toks[i]);
		    continue;
		}
		
		if (kv[0].length() < 1) {
		    System.out.println("Didn't find a clue in this token: "+toks[i]);
		    continue;
		}
		
		for (int j = 0; j < kv[0].length() ; j++) {
		    byte k = lang.charToByte(kv[0].charAt(j));
		    byte v = lang.charToByte(kv[1].charAt(j));
		    initialMap.setMapping(k, v);
		}
	    }
	}
    }

    public void scramble(boolean allowIdentityMappings)
    {
	MapSet permutedMap = new MapSet(lang.letterCount());
	permutedMap.randomlyPermute(allowIdentityMappings); 

	cipherText = lang.translateString(rawCipherText, permutedMap, null);

	// translate the clues. This requires special logic versus translateString...
	String toks[] = clues.split("[\\s,]+");

	String newclues = "";
	for (int i = 0; i < toks.length; i++) {
	    if (toks[i].length()==0)
		continue;

	    String kv[] = toks[i].split("=");
	    if (kv.length == 2) {
		newclues = newclues + lang.translateString(kv[0], permutedMap, null) + "=" + kv[1];
		if (i < toks.length-1)
		    newclues = newclues + " ";
	    }
	}

	clues = newclues;
	initialMap = new MapSet(lang.letterCount());
	initialMap.setFullSet();
	handleClues(newclues);
    }

    // remove all non-alpha letters.
    static final String onlyAlphaCharacters(String s)
    {
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (Character.isLetter(c))
		sb.append(c);
	}
	return sb.toString();
    }

    // Ensure only a single space between words (both periods and
    // spaces are considered word separators.)
    static final String onlyAlphaSpaceCharacters(String s)
    {
	StringBuffer sb = new StringBuffer();
	boolean havespace = false;

	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (Character.isLetter(c)) {
		if (havespace)
		    sb.append(' ');
		sb.append(c);
		havespace = false;
	    } else if (c==' ' || c=='.' || c=='-' || c==':' || c==';' || c=='?' || c=='!' || c=='(' || c==')') {
		havespace = true;
	    }
	}

	return sb.toString();
    }
}
