package decrypto.dictattack;

import decrypto.*;
import java.util.*;

public class DictionaryAttackSolver implements Solver
{
    ArrayList<SolverListener> listeners = new ArrayList<SolverListener>();
    double startTime = 0, endTime = 0;
    SolverThread solverThread;
    Puzzle puzzle;
    StopFlag stopFlag;
    public boolean runForever = false;

    Exception runException;

    public DictionaryAttackSolver(Puzzle puzzle)
    {
	this.puzzle = puzzle;
    }

    public Exception getException()
    {
	return runException;
    }

    public void addListener(SolverListener listener)
    {
	listeners.add(listener);
    }

    /** Asynchronously begin solving.**/
    public void start(double maxTime)
    {
	this.stopFlag = new StopFlag(maxTime);
	assert(solverThread == null);
	startTime = System.currentTimeMillis()/1000.0;
	solverThread = new SolverThread();
	solverThread.start();
    }

    /** How much time has been used for solving thus far? **/
    public double getElapsedTime()
    {
	if (endTime != 0)
	    return endTime - startTime; // we're done
	if (startTime == 0)
	    return 0;                   // haven't started

	return System.currentTimeMillis()/1000.0 - startTime;
    }

    /** Request that the solver thread stop as soon as possible. **/
    public void stop()
    {
	assert(solverThread != null);

	synchronized(solverThread) {
	    stopFlag.stop = true;
	    try {
		solverThread.join();
	    } catch (InterruptedException ex) {
		// fall through.
	    }
	}
	endTime = System.currentTimeMillis()/1000.0;
    }

    class SolverThread extends Thread
    {
	public void run()
	{
	    for (SolverListener sl : listeners)
		sl.solverMessage("Doing initial search...");
	 
	    do {
		for (Word w : puzzle.words)
		    w.enabled = true;

		SimpleDictionaryAttack sda = new SimpleDictionaryAttack(puzzle, new LazyPlanner(puzzle), listeners);
		sda.solve(stopFlag);
		if (sda.gotFullSolution && !runForever)
		    break;

		// try disabling individual words
		int iters = 1; // this is how many combinations we'll
			       // consider for this # of disabled
			       // words
		int maxDisableWordCount = Math.min(DecryptoParameters.maxRandomWordDisable, puzzle.words.size()-1);
		while (maxDisableWordCount > 1 && Math.pow(puzzle.words.size(), maxDisableWordCount)>DecryptoParameters.maxRandomWordIterations)
		    maxDisableWordCount--;

		boolean gotSolution = false;

		for (int disableWordCount = 1; disableWordCount <= maxDisableWordCount; disableWordCount++) {
		    iters *= puzzle.words.size();
		    
		    if (stopFlag.stop)
			break;

		    // try each combination of 'disableWordCount' disables...
		    for (int i = 0; i <= iters; i++) {

			// reset...
			for (Word w : puzzle.words)
			    w.enabled = true;

			String message = "Disabling words: ";

			// which words are we diabling this iteration?
			int tmp = i;
			boolean bad = false;
			for (int j = 0; j < disableWordCount; j++) {
			    int widx = tmp % puzzle.words.size();
			    tmp /= puzzle.words.size();
			    // already disabled this word, so this
			    // isn't a valid combination
			    if (!puzzle.words.get(widx).enabled)
				bad = true;
			    puzzle.words.get(widx).enabled = false;
			    message = message + puzzle.words.get(widx).text + " ";
			}

			if (!bad) {
			    for (SolverListener sl : listeners)
				sl.solverMessage(message);
			
			    sda = new SimpleDictionaryAttack(puzzle, new LazyPlanner(puzzle), listeners);
			    sda.solve(stopFlag);
			    gotSolution |= sda.gotFullSolution;
			}
		    }

		    if (gotSolution && !runForever)
			break;
		}

		for (SolverListener sl : listeners)
		    sl.solverMessage("Switching to randomized search...");
		
		int maxIters = 5000;
		
		/** We're desperate. We have several
		    non-dictionary words and we don't know what
		    they are. Consequently, we'll substitute words
		    in random order. We also must disable the
		    pruning algorithm.
		    
		    Some of these orders will be absolutely
		    miserably inefficient and so we don't want to
		    spend too much time on any particular
		    permuatation (better to try a bunch of
		    different orderings briefly than a single
		    ordering at length). Over time, however, we
		    get a little bit more patient, exploring each
		    permutation more thoroughly.
		**/
		while (!stopFlag.stop) {
		    sda = new SimpleDictionaryAttack(puzzle, new RandomPlanner(puzzle), listeners);
		    sda.maxIters = maxIters;
		    maxIters *= 1.05;
		    sda.pruneRootNode = false;
		    sda.pruneEachNode = false;
		    sda.scrambleWordOrder = true;
		    sda.solve(stopFlag);
		}
	    } while (false);

	    endTime = System.currentTimeMillis()/1000.0;
	    double elapsedTime = endTime - startTime;
		
	    for (SolverListener sl : listeners)
		sl.solverFinished(elapsedTime);

	    stopFlag.stop = true;
	}
    }
}
