package decrypto.dictattack;

import java.io.*;
import java.util.*;

import decrypto.*;

/** This class is used by DictionaryAttackSolver. In this class, we
 * perform a single recursive search. **/
public class SimpleDictionaryAttack
{
    Puzzle puzzle;
    Planner planner;

    boolean solvedWords[];
    int numSolvedWords;

    /** For each word, the index of the first valid candidate. **/
    int firstcand[];

    boolean gotFullSolution = false;

    ArrayList<SolverListener> listeners;

    public StopFlag stopFlag = null;

    public boolean pruneRootNode = true, pruneEachNode = true;

    public boolean scrambleWordOrder = false;

    int iters;
    public int maxIters = Integer.MAX_VALUE;
    public SimpleDictionaryAttack(Puzzle puzzle, Planner planner, ArrayList<SolverListener> listeners)
    {
	this.puzzle = puzzle;
	this.planner = planner;
	this.listeners = listeners;
    }

    /** truncate score to about 5 significant digits in order to
	prevent rounding errors from causing identical solutions to
	appear different **/
    static final double roundScore(double in)
    {
	long T = 1<<18;
	long s = (long) (in * T);

	return ((double) s)/T;
    }

    static final void swap(int array[], int j, int k)
    {
	int tmp = array[j];
	array[j] = array[k];
	array[k] = tmp;
    }

    static final int indexOf(int array[], int v)
    {
	for (int i = 0; i < array.length; i++)
	    if (array[i]==v)
		return i;

	return -1;
    }

    public void solve(StopFlag stopFlag)
    {
	this.stopFlag = stopFlag;
	MapSet map = puzzle.initialMap.copy();

	solvedWords = new boolean[puzzle.words.size()];
	for (int i = 0; i < puzzle.words.size(); i++)
	    puzzle.words.get(i).firstcand = 0;

	numSolvedWords = 0;

	if (pruneRootNode) {
	    boolean inconsistent = reduceCandidates(map);
	    if (inconsistent) {
		map = puzzle.initialMap.copy();
		// System.out.println("Reduced initial map is inconsistent. Resetting.");
	    }
	}

	solve(0, map);
    }

    void reportSolution(MapSet map, boolean fullSolution)
    {
	if (fullSolution)
	    gotFullSolution = true;

	double score = puzzle.lang.computeScore(puzzle.cipherBytes, map) / puzzle.numCiphertextLetters;

	for (SolverListener sl : listeners)
	    sl.solverSolution(map, null, fullSolution, score);
    }

    void solve(int recursiondepth, MapSet map)
    {
	iters++;

	if (stopFlag.stop || iters > maxIters)
	    return;

	if (numSolvedWords == puzzle.words.size()) {
	    reportSolution(map, true);
	    return;
	}
	    
	int action = planner.getAction(solvedWords, map);

	if (action == -1) {
	    reportSolution(map, true);
	    return;
	}
	
	if (action == -2) {
	    reportSolution(map, false);
	    return;
	}

	//////////////////////// LETTER ///////////////////////
	if ((action & Planner.LETTER_FLAG)!=0) {
	    byte a[] = new byte[1], b[] = new byte[1];
	    a[0] = (byte) (action & Planner.MASK);
	    
	    reportProgressSize(recursiondepth, puzzle.lang.letterCount());

	    int firstcandSave[] = new int[puzzle.words.size()];
	    saveFirstcandState(firstcandSave);

	    MapSet ourmap = map.copy();

	    // do we have any children?
	    boolean leaf = true;

	    for (int c = 0; c < puzzle.lang.letterCount(); c++) {
		reportProgress(recursiondepth, c);
		    
		b[0] = (byte) c;
		    
		if (ourmap.isMappingOkay(a, b)) { 

		    ourmap.setMappings(a, b);
			    
		    // We MUST reduce candidates when substituting a single letter.
		    // That's the whole point...
		    // if (DecryptoParameters.pruneEachNode)
		    boolean inconsistent = reduceCandidates(ourmap); // don't increment wordnum
		    
		    if (!inconsistent) {
			solve(recursiondepth+1, ourmap);
			leaf = false;
		    }

		    ourmap.setTo(map);
		    restoreFirstcandState(firstcandSave);
		}
	    }

	    if (leaf)
		reportSolution(map, false);

	    return;
	}

	//////////////////////// WORD ///////////////////////
	if ((action & Planner.WORD_FLAG)!=0) {
	    int wordnum = action & Planner.MASK;

	    Word w = puzzle.words.get(wordnum);

	    solvedWords[wordnum] = true;
	    numSolvedWords++;

	    if (!w.enabled) {
		assert(false);
		solve(recursiondepth+1, map);
	    } else {
		// this can be accelerated. See SpaceDecryptoSolver
		int firstcandSave[] = new int[puzzle.words.size()];
		saveFirstcandState(firstcandSave);

		MapSet ourmap = map.copy();
	    
		reportProgressSize(recursiondepth, w.candidates.length - w.firstcand);

		// do we have any children?
		boolean leaf = true;

		if (scrambleWordOrder) {
		    Random r = new Random();
		    for (int c = w.firstcand; c < w.candidates.length; c++) {
			int d = r.nextInt(w.candidates.length - w.firstcand) + w.firstcand;
			Candidate cc = w.candidates[c];
			Candidate cd = w.candidates[d];
			w.candidates[d] = cc;
			w.candidates[c] = cd;
		    }
		}

		for (int c = w.firstcand; c < w.candidates.length; c++) {

		    reportProgress(recursiondepth, c - w.firstcand);
		    
		    Candidate candidate = w.candidates[c];
		    
		    // if this word can't map successfully, try the next word.
		    if (!ourmap.isMappingOkay(w.word, candidate.cand))
			continue;
		    
		    // we found a match; we're not a leaf anymore.
		    leaf = false;
		    
		    // assume that this word maps to candidate.
		    ourmap.setMappings(w.word, candidate.cand);

		    boolean inconsistent = false;
		    if (pruneEachNode)
			inconsistent = reduceCandidates(ourmap);

		    if (!inconsistent) {
			solve(recursiondepth+1, ourmap);
			leaf = false;
		    }

		    // we tampered with our map; restore it back to its
		    // original configuration so we can try a different mapping
		    ourmap.setTo(map);
		    restoreFirstcandState(firstcandSave);
		}
	    
		/* we're at a leaf. We report partial solutions if either
		   they've requested it, or if we've already solved for a lot
		   of words. */
		if (leaf)
		    reportSolution(ourmap, false);
	    }

	    numSolvedWords--;
	    solvedWords[wordnum] = false;
	    return;
	}
	System.out.println("Planner error: No action specified (need WORD or LETTER flag).");
    }

    protected boolean reduceCandidates(MapSet set)
    {
	MapSet newset = new MapSet(puzzle.lang.letterCount());

	int iters = 0;

	boolean dirty = true;

	while (dirty && !stopFlag.stop && iters < DecryptoParameters.maxReduceIterations) {

	    dirty = false;

	    // take a look at every not-yet-solved cipherword
	    for (int wordnum = 0; wordnum < puzzle.words.size(); wordnum++) {

		if (solvedWords[wordnum])
		    continue;

		Word w = puzzle.words.get(wordnum);
		if (!w.enabled)
		    continue;

		newset.setEmptySet();

		for (int c = w.firstcand; c < w.candidates.length; c++) {

		    Candidate candidate = w.candidates[c];
				 
		    if (set.isMappingOkay(w.word, candidate.cand)) {

			newset.enableMappings(w.word, candidate.cand);

		    } else {

			// shuffle this candidate to
			// index "firstcand", then
			// increase "firstcand". We've
			// already processed the
			// element at "firstcand", so
			// we don't have to worry
			// about missing it.
			Candidate tmp = w.candidates[w.firstcand];
			w.candidates[w.firstcand]=candidate;
			w.candidates[c] = tmp;
			w.firstcand++;

			dirty = true;
		    }
		}

		set.intersectWith(newset);
	    }

	    // can we reduce the set by any other logic?
	    // i.e., if X={B,C} and Y={B,C}, then no other letter can map to B or C.
	    if (dirty)
		set.selfReduce();


	    // Have we become inconsistent? 
	    // A) do any letters have no mappings left?
	    // B) do any two letters map to the same plain text?
	    long mappedTo = 0;
	    for (int i = 0; i < puzzle.lang.letterCount(); i++) {
		if (!set.hasMappings((byte) i))
		    return true;

		int mi = set.getMapping((byte) i);
		if (mi >= 0) {
		    if ((mappedTo & (1<<mi))>0)
			return true;
		    mappedTo |= (1<<mi);
		}
	    }

	    iters++;
	}

	// not inconsistent (that we can tell)
	return false;
    }

    int reportunits = 0;
    int reportdepth = 2;;
    static final int MAXREPORTDEPTH = 10;

    // how much work is there to do at recursion level N?
    int worksize[] = new int[MAXREPORTDEPTH]; 

    // how much work ahve we completed at recursion level N?
    int workprogress[] = new int[MAXREPORTDEPTH];

    double lastprogress = -1;

    void reportProgressSize(int depth, int amount)
    {
	/*
	if (reportunits < 1024 && depth>reportdepth && depth < MAXREPORTDEPTH) {
	    reportdepth = depth;
	    if (reportunits == 0)
		reportunits = amount;
	    else
		reportunits *= amount;
	}
	*/

	if (depth <= reportdepth)
	    worksize[depth] = amount;
    }
    
    void reportProgress(int depth, int amount)
    {
	if (depth <= reportdepth) {
	    workprogress[depth] = amount;
	    
	    double frac = 1.0;
	    double prog = 0;
	    for (int i = 0; i <= depth; i++) {
		prog += (frac*workprogress[i])/worksize[i];
		frac /= worksize[i];
	    }

	    if (prog - lastprogress < 0.03)
		return;

	    lastprogress = prog;

	    for (SolverListener listener : listeners)
		listener.solverProgress(prog);
	}
    }

    void saveFirstcandState(int v[])
    {
	for (int widx = 0; widx < puzzle.words.size(); widx++)
	    v[widx] = puzzle.words.get(widx).firstcand;
    }

    void restoreFirstcandState(int v[])
    {
	for (int widx = 0; widx < puzzle.words.size(); widx++)
	    puzzle.words.get(widx).firstcand = v[widx];
    }
}

