package decrypto.genetic;

import java.util.*;
import java.io.*;

import decrypto.*;
import decrypto.statistics.*;

/////////////////////////////////////////////////////////////////
// Information that is constant for all the genetic solvers used
// to solve a particular ciphertext.
public class GAProblem
{
    Puzzle           puzzle;

    // Suppose the mappings for two ciphertext letters 'a' and 'b'
    // have been swapped. table[a][b] is a list of the indices into
    // the cipherbytes where the affected trigrams can be found.
    short            table[][][];
	 
    int              nextGenerationId;

    Random           rand = new Random();
    double           scorethresh = -Double.MAX_VALUE;

    /*
    float            trigrams[][][];
    short            strigrams[][][];
    */

    float            tetragrams[][][][];
    float            tetragramsNS[][][][];

    // When scoring a puzzle, we wrap the end of the puzzle around
    // back to the beginning. E.g., for 4-grams, "the quick brown fox"
    // becomes "ox the quick brown fox". This allows the 4-grams for
    // the first character to be meaningful.
    //    byte            scoringCipherBytes[];

    public GAProblem(Puzzle puzzle)
    {
	this.puzzle = puzzle;
	this.table = precomputeIndicesTable();

	/*
	if (puzzle.trustSpaces)
	    trigrams = (float[][][]) puzzle.lang.getNgrams(3, true);
	else
	    trigrams = (float[][][]) puzzle.lang.getNgrams(3, false);
	*/

	tetragrams = (float[][][][]) puzzle.lang.getNgrams(4, true);
	tetragramsNS = (float[][][][]) puzzle.lang.getNgrams(4, false);

	//	makeShortTrigrams();

	// +1 to allow for SPACE character
	if (false) {
	    int identitymap[] = new int[puzzle.lang.letterCount()+1];
	    for (int i = 0; i < identitymap.length; i++)
		identitymap[i] = i;
	    System.out.printf("initial score: %15.5f\n", computeScore(identitymap)/puzzle.numCiphertextLetters);
	}
    }

    // for every combination of two ciphertext symbols, compute the
    // set of trigrams that are affected by those symbols. The
    // trigrams are represented by the index of the *first* character
    // in the trigram.
    short[][][] precomputeIndicesTable()
    {
	// some ngrams are duplicated in the ciphertext. This hashmap
	// gives the lowest #'d ngram for the ngram starting at a
	// given idx.
	HashMap<String, Integer> canonicalMap = new HashMap<String, Integer>();
	int canonicalNgrams[] = new int[puzzle.cipherBytes.length];

	// broken for puzzles that contain spaces? No. Even though
	// spaces will be transformed into a non-alphabetic character, they are 
	// transformed in the same way each time. Thus we'll still work.
	for (int i = 0; i < puzzle.cipherBytes.length-3; i++) {
	    String s = ""+((char) puzzle.cipherBytes[i] + 'A') + 
		((char) puzzle.cipherBytes[i+1] + 'A') + 
		((char) puzzle.cipherBytes[i+2] + 'A') +
		((char) puzzle.cipherBytes[i+3] + 'A');

	    Integer canonical = canonicalMap.get(s);
	    if (canonical == null) {
		canonicalMap.put(s, i);
		canonical = i;
	    }
	    canonicalNgrams[i] = canonical;
	}

	// Suppose we swap the plaintext mappings for two
	// letters. Changing these two mappings will affect only a
	// subset of the ngrams in the puzzle. The twolettertable
	// forms a list of all the ngrams that are affected when two
	// mappings are swapped.
	short twolettertable[][][] = new short[26][26][];
	    
	for (int a = 0; a < 26; a++) {
	    for (int b = 0; b < 26; b++) {
		ArrayList<Integer> indices = new ArrayList<Integer>();
		    
		for (int i = 0; i < puzzle.cipherBytes.length-3; i++) {
		    // would the trigram beginning at index 'i' be affected by a change
		    // to ciphertext symbol 'a' or 'b'?
		    if (puzzle.cipherBytes[i]==a || puzzle.cipherBytes[i+1]==a || puzzle.cipherBytes[i+2]==a ||
			puzzle.cipherBytes[i]==b || puzzle.cipherBytes[i+1]==b || puzzle.cipherBytes[i+2]==b)
			indices.add(canonicalNgrams[i]);
		}

		// sort the canonical indices. This will put equivalent n-grams next to each other,
		// e.g., 1, 2, 3, 3, 3, 4, 5, 6, 7, 7, 8, ...
		Collections.sort(indices);

		// Now do RLE compression on those indices
		ArrayList<Integer> compressed = new ArrayList<Integer>();
		int idx = 0;
		while (idx < indices.size()) {
		    // here's the value
		    int v = (short) (int) indices.get(idx);

		    // compute the run length
		    int eidx = idx;
		    while (eidx+1 < indices.size() && indices.get(eidx+1)==v)
			eidx++;
			
		    // emit the (value, length) pair
		    compressed.add(v);
		    compressed.add(eidx-idx+1);
		    idx = eidx+1;
		}
		    
		// convert the output into an array of shorts.
		twolettertable[a][b] = new short[compressed.size()];
		for (int i = 0; i < compressed.size(); i++)
		    twolettertable[a][b][i] = (short) (int) compressed.get(i);
	    }
	}
	    
	return twolettertable;
    }

    public final double computeScore(int map[])
    {
	double score = 0;

	// naively, we have the problem that we have no good way to
	// bootstrap the N-th order predictor for the first few
	// characters. Our approach here is to pretend that the
	// cipherBytes form a loop.

	int ca = puzzle.lang.letterCount(), cb = ca, cc = ca; // our preceeding characters

	int len = puzzle.cipherBytes.length;
	float grams[][][][];

	if (puzzle.trustSpaces) {
	    // a loop with a space.
	    if (len > 2)
		ca = puzzle.cipherBytes[len-2];
	    if (len > 1)
		cb = puzzle.cipherBytes[len-1];
	    cc = puzzle.lang.letterCount();
	    grams = tetragrams;
	} else {
	    // a loop without a space
	    if (len > 3)
		ca = puzzle.cipherBytes[len-3];
	    if (len > 2)
		cb = puzzle.cipherBytes[len-2];
	    if (len > 1)
		cc = puzzle.cipherBytes[len-1];
	    grams = tetragramsNS;
	}

	int ma, mb, mc;
	ma = map[ca];
	mb = map[cb];
	mc = map[cc];

	for (int idx = 0; idx < puzzle.cipherBytes.length; idx++) {
	    int cd = puzzle.cipherBytes[idx];
	    int md = map[cd];

	    score += grams[ma][mb][mc][md];
	    //	    System.out.printf("%3d %15f %c%c%c%c %15f\n", idx, grams[ma][mb][mc][md], ma+'A', mb+'A', mc+'A', md+'A', score);

	    ma = mb;
	    mb = mc;
	    mc = md;
	}
	    
	return score;
    }
}
