package decrypto.genetic;

import java.util.*;
import java.io.*;

import decrypto.*;
import decrypto.statistics.*;

// A Patristocrat solver, i.e., a simple substitution cipher in which
// spaces have been omitted.  (It can also respect word boundaries,
// thus solving cryptograms as well.)
//
// The patristocrat solving is based on finding the ML sequence of
// characters based on a corpus of text in which spaces have been
// removed. Spaces are then inserted such that the resulting trigram
// probabilities are maximized. The implementation here produces the
// optimal trigram-based space estimates, but these estimates are
// often disappointing.

public class GeneticSearch
{
    /////////////////////////////////////////////////////////////////
    // State of this solver
    GAProblem problem;
    Puzzle    puzzle;

    int    map[];                              // current state of solver (best so far)
    double score = -Double.MAX_VALUE;          // log-probability of state

    int    totaliterations = 0;                // how many iterations have we performed?

    ArrayList<SolverListener> listeners = new ArrayList<SolverListener>();
    StopFlag stopFlag;

    class LetterChoice implements Comparable<LetterChoice>
    {
	int cipherletter;
	int nchoices;

	public int compareTo(LetterChoice lc)
	{
	    return nchoices - lc.nchoices;
	}
    }

    public GeneticSearch(GAProblem problem, StopFlag stopFlag)
    {
	this.problem = problem;
	this.puzzle = problem.puzzle;
	this.stopFlag = stopFlag;

	makeInitialMap();
    }

    boolean makeInitialMap()
    {
	///////////////////////////////////////////
	// initialize a new random solution
	// +1 to account for space at the end.
	map = new int[puzzle.lang.letterCount()+1];
	map[puzzle.lang.letterCount()] = puzzle.lang.letterCount();

	// We randomly generate initial solutions. To do this
	// efficiently, we start with the most constrained letters
	// first.
	ArrayList<LetterChoice> choices = new ArrayList<LetterChoice>();
	for (int i = 0; i < puzzle.lang.letterCount(); i++) {
	    LetterChoice lc = new LetterChoice();
	    lc.cipherletter = i;
	    lc.nchoices = 0;
	    choices.add(lc);

	    // how many letters can cipherletter i map to?
	    for (int j = 0; j < puzzle.lang.letterCount(); j++)
		if (puzzle.initialMap.isMappingOkay((byte) i, (byte) j))
		    lc.nchoices++;
	}

	// sort with most-constrained letters first. i.e., choices.get(0).nchoices is the minimum.
	Collections.sort(choices);

	// Now, generate the words.
	boolean badinit = true;
	int trials = 0;

	while (badinit) {

	    synchronized(stopFlag) {
		if (stopFlag.stop) {
		    map = null;
		    return false;
		}
	    }

	    boolean mapped[] = new boolean[puzzle.lang.letterCount()];
	    badinit = false;

	    for (int i = 0; i < puzzle.lang.letterCount(); i++) {

		int ci = choices.get(i).cipherletter;

		// which plaintext letters could ci legally map to?
		int possibilities[] = new int[puzzle.lang.letterCount()];
		int npossibilities = 0;
		
		for (int j = 0; j < puzzle.lang.letterCount(); j++) {
		    if (puzzle.initialMap.isMappingOkay((byte) ci, (byte) j) && !mapped[j]) {
			possibilities[npossibilities++] = j;
		    }
		}

		if (npossibilities == 0) {
		    badinit = true;
		    break;
		}
		
		// pick one of the possibilities.
		int pi = possibilities[problem.rand.nextInt(npossibilities)];
		map[ci] = pi;
		mapped[pi] = true;
	    }
	}

	return true;
    }

    void reportSolution(int map[], double score)
    {
	int spaces[] = null;
	MapSet mapset = new MapSet(map);
	
	if (!puzzle.trustSpaces) {
	    spaces = findSpaces4(puzzle, puzzle.cipherBytes, map);
	}
	
	for (SolverListener listener : listeners) {
	    listener.solverSolution(mapset, spaces, false, score / puzzle.numCiphertextLetters);
	}
    }

    void shuffle(int v[])
    {
	for (int i = 0; i < v.length; i++) {
	    int j = problem.rand.nextInt(v.length-i)+i;
	    int mi = v[i];
	    int mj = v[j];
	    v[i] = mj;
	    v[j] = mi;
	}
    }

    // pick npermutes ciphertext letters and their plaintext mapping,
    // and randomly shuffle around their mappings. We keep shuffling
    // until we find a mapping that respects the initialMap.
    //
    // May return null if no valid mutation can be found.
    public GeneticSearch mutate(int npermutes)
    {
	GeneticSearch gas = new GeneticSearch(problem, stopFlag);

	// start with our current map.
	if (map == null || gas.map == null)
	    return null;

	System.arraycopy(map, 0, gas.map, 0, map.length);

	int cletters[] = new int[npermutes];
	int pletters[] = new int[npermutes];
	boolean clettersPicked[] = new boolean[puzzle.lang.letterCount()];

	for (int i = 0; i < npermutes; i++) {
	    int idx;

	    // pick a ciphertext letter.
	    do {
		idx = problem.rand.nextInt(puzzle.lang.letterCount());
	    } while (clettersPicked[idx]);
		
	    clettersPicked[idx] = true;

	    // add it to our list of letters to permute.
	    cletters[i] = idx;
	    pletters[i] = map[idx];
	}

	// find an acceptable shuffling.
	int trials = 0;

	while (true) {

	    shuffle(cletters);
	    
	    boolean bad = false;
	    
	    for (int i = 0; i < cletters.length; i++) {
		if (!puzzle.initialMap.isMappingOkay((byte) cletters[i], (byte) pletters[i])) {
		    bad = true;
		    break;
		}
	    }
	    
	    if (!bad)
		break;

	    trials++;
	    if (trials > 1000)
		return null;
	}

	for (int i = 0; i < cletters.length; i++)
	    gas.map[cletters[i]] = pletters[i];
	
	return gas;
    }

    int[] makeSwaps()
    {
	int swaps[] = new int[puzzle.lang.letterCount() * (puzzle.lang.letterCount()-1)];
	int swappos = 0;

	for (int i = 0; i < puzzle.lang.letterCount(); i++) {
	    for (int j = 0; j < puzzle.lang.letterCount(); j++) {
		if (i==j)
		    continue;

		swaps[swappos++] = (i<<16)+j;
	    }
	}

	return swaps;
    }

    public void iterate(int maxIterations, boolean protect[])
    {
	if (map == null) // didn't get a good init
	    return;

	long starttime = System.nanoTime();
	int iters = 0;
	int swaps[] = makeSwaps();

	score = problem.computeScore(map);

	if (score > problem.scorethresh)
	    reportSolution(map, score);

	for (iters = 0; iters < maxIterations; iters++, totaliterations++) {

	    boolean keptSwap = false;
	    shuffle(swaps);

	    for (int swapidx = 0; swapidx < swaps.length; swapidx++) {

		// pick two cipher characters to swap
		int a = swaps[swapidx]>>16;
		int b = swaps[swapidx]&0xffff;
		double newscore = 0;

		if (!puzzle.initialMap.isMappingOkay((byte) a, (byte) map[b]) ||
		    !puzzle.initialMap.isMappingOkay((byte) b, (byte) map[a]))
		    continue;
		
		if (protect != null && (protect[a] || protect[b]))
		    continue;

		// swap the chars
		int ma = map[a], mb = map[b];
		map[a] = mb; map[b] = ma;
		
		newscore = problem.computeScore(map);
		
		if (newscore <= score) {
		    // if it got worse (or no change), then undo the change
		    map[a] = ma; map[b] = mb;
		    continue;
		}

		// we made an improvement, we'll keep this change.
		score = newscore;
		keptSwap = true;

		// should we report the solution?
		if (score > problem.scorethresh) {
		    reportSolution(map, score);
		}
	    }

	    // if local minimum, exit.
	    if (!keptSwap)
		break;
	}
	
	// System.out.printf("%10d %10d\n", iters, maxIterations);

	long endtime = System.nanoTime();
	if (false) {
	    double dt = (endtime - starttime)/1000000000.0;
	    System.out.printf("ips: %15.5f\n", maxIterations/dt);
	}
    }

   /** Compute the maximum likelihood position of spaces given a
        particular plaintext sequence, using trigram probabilities.
        This method is linear time in the number of characters (very
        fast!), based on a forward viterbi method. It is also
        *optimal* (in terms of maximizing the tetragram probabilities).

        That said, the performance often leaves something to desired.
    **/
    static int[] findSpaces4(Puzzle puzzle, byte cipherbytes[], int map[])
    {
        // tetragram model == third order markov model. In forward
        // viterbi algorithm, we have eight possible states at each
        // node. We care about the last three characters.
        //
        // 0: abc
        // 1: abc_
        // 2: ab_c
        // 3: ab_c_
        // 4: a_bc
        // 5: a_bc_
        // 6: a_b_c
        // 7: a_b_c_
        //
        // where a, b, c are the previous cipherbytes characters.
        //
        int spaces[][] = new int[8][cipherbytes.length];
        double scores[] = new double[8];

	if (cipherbytes.length<4)
	    return spaces[0];

	float grams[][][][] = (float[][][][]) puzzle.lang.getNgrams(4, true);

        // initialize scores, taking into account spaces[0] and spaces[1]
        if (true) {

            int tsv[] = new int[8];
            int tsvlen;

            for (int j = 0; j < 8; j++) {
                tsvlen = 0;

		// the last character of the plaintext
		tsv[tsvlen++] = map[cipherbytes[cipherbytes.length-1]];

		// add a space
                tsv[tsvlen++] = 26;

		// the first character of the plaintext
                tsv[tsvlen++] = map[cipherbytes[0]];

		// followed by a space (for some of our sequences...)
                if ((j&4)>0)
                    tsv[tsvlen++] = 26;

		// the second character of the plaintext
                tsv[tsvlen++] = map[cipherbytes[1]];

		// followed by a space (for some of our sequences...)
                if ((j&2)>0)
                    tsv[tsvlen++] = 26;

		// the third character of the plaintext
                tsv[tsvlen++] = map[cipherbytes[2]];

		// followed by a space (for some of our sequences...)
                if ((j&2)>0)
                    tsv[tsvlen++] = 26;

		// now compute the probability of this particular sequence.
                for (int k = 0; k < tsvlen-3; k++)
                    scores[j] += grams[tsv[k]][tsv[k+1]][tsv[k+2]][tsv[k+3]];
            }
        }

        int newspaces[][] = new int[8][cipherbytes.length];

        int tsv[] = new int[8];
        int tsvlen;

        // each node will be (temporarily) expanded into two new nodes
        double expandedscores[] = new double[16];

        for (int i = 3; i < cipherbytes.length; i++) {

            // expand each of the four nodes into eight nodes
            for (int j = 0; j < 8; j++) {

                tsvlen = 0;

                tsv[tsvlen++] = map[cipherbytes[i-3]];
                if ((j&4)>0)
                    tsv[tsvlen++] = 26;
                tsv[tsvlen++] = map[cipherbytes[i-2]];
                if ((j&2)>0)
                    tsv[tsvlen++] = 26;
                tsv[tsvlen++] = map[cipherbytes[i-1]];
                if ((j&1)>0)
                    tsv[tsvlen++] = 26;
                tsv[tsvlen++] = map[cipherbytes[i]];

                int md = tsv[tsvlen-1], mc = tsv[tsvlen-2], mb = tsv[tsvlen-3], ma = tsv[tsvlen-4];

                // spaces[i] = 0
                expandedscores[j*2+0] = scores[j] + grams[ma][mb][mc][md];

		// every sequence ends in a space.
		if (i == cipherbytes.length-1)
		    expandedscores[j*2+0] += grams[mb][mc][md][26];

                // spaces[i] = 1
                expandedscores[j*2+1] = scores[j] + grams[ma][mb][mc][md] + grams[mb][mc][md][26];
            }

            // The value of space[i-2] (A) will not affect any future
            // grams, whereas the values of space[i-1] (B) and
            // space[i] (C) will.
            //
            // Thus, for each possible value of BC, we will pick the
            // value of A (and all preceeding values) that maximizes
            // the probability. For example, ABC=000 and ABC=100 will
            // be collapsed into a new node BC=00, which will inherit the
            // set of spaces.
            //

	    // (NB: this is the diagram for trigrams)

            // original  expanded   collapse pairs 
            //    ab       abc                          bc
            //    00       000      *------------------ 00
            //             001       |  *-------------- 01
            //    01       010       |   |  *---------- 10
            //             011       |   |   |  *------ 11
            //    10       100      *-   |   |   |        
            //             101          *-   |   |
            //    11       110              *-   |        
            //             111                  *-

	    //   abc    expanded    pairs        bcd
	    //   ------------------------------------------
	    //   000     0000         *          000      
	    //           0001         |*         001            
	    //   001     0010         ||*        010      
	    //           0011         |||*       011      
	    //   010     0100         ||||*      100                                   
	    //           0101         |||||*     101                                   
	    //   011     0110         ||||||*    110      
	    //           0111         |||||||*   111      
	    //   100     1000         *|||||||            
	    //           1001          *||||||            
	    //   101     1010           *|||||                                         
	    //           1011            *||||             
	    //   110     1100             *|||            
	    //           1101              *||            
	    //   111     1110               *|            
	    //           1111                *            
            // collapse nodes, deciding on the value of spaces[j][i-2]

            // XXX: This implementation is not O(N) because we are
            // copying the entire history (the loop in k) rather than
            // "linking" to the previous solution.
            for (int j = 0; j < 8; j++) {
		// NB: j is looping over values of BC.
		// we decide between expanded[j] and expanded[j+8]
                if (expandedscores[j] >= expandedscores[j+8]) {
                    // inherit the solution from original j/2
                    for (int k = 0; k < i-3; k++)
                        newspaces[j][k] = spaces[j/2][k];
                    // we've decided A
                    newspaces[j][i-3] = 0;
                    scores[j] = expandedscores[j];
                } else {
                    // inherit the solution
                    for (int k = 0; k < i-3; k++)
                        newspaces[j][k] = spaces[(j+8)/2][k];
                    // we've decided A
                    newspaces[j][i-3] = 1;
                    scores[j] = expandedscores[j+8];
                }
            }
            int tmp[][] = spaces;
            spaces = newspaces;
            newspaces = tmp;
        }

        // now return the set of spaces with the best score
        int bestidx = -1;
        double bestscore = -Double.MAX_VALUE;

        for (int j = 0; j < 8; j++) {

            // decide on last two values of spaces
            spaces[j][cipherbytes.length-3] = (j&4)>0 ? 1 : 0;
            spaces[j][cipherbytes.length-2] = (j&2)>0 ? 1 : 0;
            spaces[j][cipherbytes.length-1] = (j&1)>0 ? 1 : 0;

            if (scores[j] > bestscore) {
                bestscore = scores[j];
                bestidx = j;
            }
        }

        // this removes an extra space from being displayed at the end
        // of the string. This has no semantic function, as the
        // answers are indistinguishable to humans. 
        spaces[bestidx][cipherbytes.length-1] = 0;

        return spaces[bestidx];
    }

    public static void main(String args[])
    {
	try {
	    mainEx(args);
	} catch (IOException ex) {
	}
    }

    static void mainEx(String args[]) throws IOException
    {
	Dict dict = new Dict("english-standard.dat");
	Puzzle p = new Puzzle(dict, args[0], args[1], false, true);

	int map[] = new int[p.lang.letterCount()+1];
	for (int i = 0; i < map.length; i++)
	    map[i] = i;

	int spaces[] = findSpaces4(p, p.cipherBytes, map);
	System.out.println(p.lang.translateString(p.cipherText, new MapSet(map), spaces));
    }
}
