package decrypto.genetic;

import java.util.*;
import java.io.*;

import decrypto.*;
import decrypto.statistics.*;

// A Patristocrat solver, i.e., a simple substitution cipher in which
// spaces have been omitted.  (It can also respect word boundaries,
// thus solving cryptograms as well.)
//
// The patristocrat solving is based on finding the ML sequence of
// characters based on a corpus of text in which spaces have been
// removed. Spaces are then inserted such that the resulting trigram
// probabilities are maximized. The implementation here produces the
// optimal trigram-based space estimates, but these estimates are
// often disappointing.

public class GeneticSolver implements Solver
{
    GAProblem problem;
    Puzzle    puzzle;
    ArrayList<SolverListener> listeners = new ArrayList<SolverListener>();
    SolverThread solverThread;
    double startTime = 0, endTime = 0;
    StopFlag stopFlag;

    static HashMap<String, Sampler<String>> gramfreqws; // with spaces
    static HashMap<String, Sampler<String>> gramfreqns; // no spaces

    HashMap<String, Sampler<String>> gramfreq;

    Exception runException;

    static {
	try {
	    gramfreqns = decrypto.statistics.GramFreq.readBinary("gramfreq5ns.stb");
	    gramfreqws = decrypto.statistics.GramFreq.readBinary("gramfreq5.stb");
	} catch (Exception ex) {
	    System.out.println("ex: "+ex);
	    ex.printStackTrace();
	}
    }

    public GeneticSolver(GAProblem problem)
    {
	this.problem = problem;
	this.puzzle = problem.puzzle;

	if (puzzle.trustSpaces)
	    gramfreq = gramfreqws;
	else
	    gramfreq = gramfreqns;
    }

    public Exception getException()
    {
	return runException;
    }

    public void addListener(SolverListener listener)
    {
	listeners.add(listener);
    }

    public void start(double maxTime)
    {
	this.stopFlag = new StopFlag(maxTime);
	assert(solverThread == null);
	startTime = System.currentTimeMillis()/1000.0;
	solverThread = new SolverThread();
	solverThread.start();
    }

    public double getElapsedTime()
    {
	if (endTime != 0)
	    return endTime - startTime; // we're done
	if (startTime == 0)
	    return 0;                   // haven't started

	return System.currentTimeMillis()/1000.0 - startTime;
    }

    public void stop()
    {
 	assert(solverThread != null);

	if (Thread.currentThread()!=solverThread) {

	    synchronized(solverThread) {
		stopFlag.stop = true;
		try {
		    solverThread.join();
		} catch (InterruptedException ex) {
		    // fall through.
		}
	    }
	} else {
	    stopFlag.stop = true;
	}

	endTime = System.currentTimeMillis()/1000.0;
    }

    // shuffle the first vlen elements of v.
    static final void shuffle(Random r, byte v[], int vlen)
    {
	for (int i = 0; i < vlen; i++) {
	    int j = r.nextInt(vlen-i)+i;
	    byte mi = v[i];
	    byte mj = v[j];
	    v[i] = mj;
	    v[j] = mi;
	}
    }


    /** multiSolve is a wrapper around GASolver that does two things:
     * first, it attempts to avoid local minima by periodically
     * starting new searches. Second, it detects convergence by
     * testing when the periodically restarted solvers find the same
     * solution as the current best solution. 
     *
     * In general, the problem is that we don't know a priori how long
     * any one solver should be left to run. Consequently, we increase
     * the run time for any particular search by a fixed percentage
     * after each search.
     **/ 
    class SolverThread extends Thread
    {
	GeneticSearch makeMutatedSearch(GeneticSearch bestSearch)
	{
	    GeneticSearch newSearch = bestSearch.mutate(1+problem.rand.nextInt(puzzle.lang.letterCount()));
	    if (newSearch == null)
		return new GeneticSearch(problem, stopFlag);

	    // guess a set of matchings based on n-gram probabilities. First,
	    // randomly pick a section of ciphertext to map.
	    int len = gramfreq.keySet().iterator().next().length(); 

	    if (puzzle.cipherBytes.length <= len)
		return null;

	    int offset = problem.rand.nextInt(puzzle.cipherBytes.length - len);

	    byte ciphersnippet[] = new byte[len];
	    
	    // copy selected ciphertext... ciphersnippet might contain spaces, duplicated letters!
	    for (int i = 0; i < len; i++)
		ciphersnippet[i] = puzzle.cipherBytes[offset+i];
	    
	    // Based on the pattern of ciphertext letters, sample from the distribution
	    // of possible plaintext.
	    String pattern = decrypto.statistics.GramFreq.makePattern(puzzle.lang.bytesToString(ciphersnippet));
	    Sampler<String> sampler = gramfreq.get(pattern);
	    if (sampler == null)
		return null;

	    String plainsnippetstring = sampler.sample(problem.rand.nextDouble());
	    
	    // Make a list of ciphertext letters -> plaintext letters that we want to establish.
	    // (We have to filter out spaces.)
	    byte cipherbytes[] = new byte[len];
	    byte plainbytes[]  = new byte[len];
	    int  ncipherbytes = 0;

	    for (int i = 0; i < ciphersnippet.length; i++) {
		byte cipherbyte = ciphersnippet[i];
		byte plainbyte = puzzle.lang.charToByte(plainsnippetstring.charAt(i));
		
		// skip spaces.
		if (plainbyte == puzzle.lang.letterCount())
		    continue;

		// skip duplicates.
		boolean alreadyhave = false;
		for (int j = 0; j < i; j++)
		    if (cipherbytes[j]==cipherbyte)
			alreadyhave = true;

		if (!alreadyhave) {
		    cipherbytes[ncipherbytes] = cipherbyte;
		    plainbytes[ncipherbytes] = plainbyte;
		    ncipherbytes++;

		    if (!puzzle.initialMap.isMappingOkay(cipherbyte, plainbyte))
			return null;
		}
	    }

	    // Establishing these cipherbytes->plainbytes mappings will displace some other
	    // ciphertext and plaintext letters.
	    byte cipherbyteorphans[] = new byte[len];
	    int ncipherbyteorphans = 0;
	    byte plainbyteorphans[] = new byte[len];
	    int nplainbyteorphans = 0;

	    for (int cbidx = 0; cbidx < ncipherbytes; cbidx++) {
		// some other ciphertext letter used to map to
		// plainbytes[i]. Which one is it?

		// which ciphertext letter (a) currently maps to pcs[i]?
		byte cipherbyteorphan = -1;
		for (byte j = 0; j < puzzle.lang.letterCount(); j++) {
		    if (newSearch.map[j] == plainbytes[cbidx]) {
			cipherbyteorphan = j;
			break;
		    }
		}

		// However, if cipherbyteorphan gets a new assignment from cipherbytes, it's NOT an orphan.
		if (true) {
		    boolean orphan = true;

		    for (int i = 0; i < ncipherbytes; i++) {
			if (cipherbytes[i] == cipherbyteorphan)
			    orphan = false;
		    }

		    if (orphan) 
			cipherbyteorphans[ncipherbyteorphans++] = cipherbyteorphan;
		}

		// The plaintext letter that cipherbyte[i] used to map
		// to might be an orphan.
		byte plainbyteorphan = (byte) newSearch.map[cipherbytes[cbidx]];

		// But it's not an orphan if our mapping gives it a new owner.
		if (true) {
		    boolean orphan = true;

		    for (int i = 0; i < ncipherbytes; i++) {
			if (plainbytes[i] == plainbyteorphan) 
			    orphan = false;
		    }

		    if (orphan)
			plainbyteorphans[nplainbyteorphans++] = plainbyteorphan;
		}
	    }

	    // now, scramble the assignments of orphans.
	    assert(ncipherbyteorphans == nplainbyteorphans);

	    int shuffletrials = 0;
	    while (true) {
		shuffle(problem.rand, cipherbyteorphans, ncipherbyteorphans);
		
		boolean bad = false;
		for (int i = 0; i < ncipherbyteorphans; i++)
		    if (!puzzle.initialMap.isMappingOkay(cipherbyteorphans[i], plainbyteorphans[i]))
			bad = true;

		if (!bad)
		    break;

		shuffletrials++;
		if (shuffletrials > 100)
		    return null;
	    }
	    
	    // now make those assignments!
	    for (int i = 0; i < ncipherbyteorphans; i++) 
		newSearch.map[cipherbyteorphans[i]] = plainbyteorphans[i];
	    for (int i = 0; i < ncipherbytes; i++)
		newSearch.map[cipherbytes[i]] = plainbytes[i];

	    return newSearch;
	}

	public void run()
	{
	    try {
		runEx();
	    } catch (Exception ex) {
		runException = ex;
		ex.printStackTrace();
	    }
	}

	void runEx()
	{
	    for (SolverListener listener : listeners)
		listener.solverMessage("Searching...");

	    ArrayList<GeneticSearch> bestSearches = new ArrayList<GeneticSearch>();

	    int loops = 0;
	    while (true) {
		loops ++;

		synchronized(this) {
		    if (stopFlag.stop)
			break;
		}
		/*
		if (noprogress == 15000) {
		    bestSearch = new GeneticSearch(problem, stopFlag);
		    //  System.out.println("BOOM "+loops);
		    noprogress = 0;
		}
		*/

		// create a brand new solver and iterate it.
		GeneticSearch newSearch = null; 
		boolean protect[] = null; //(problem.rand.nextInt(2)==0 ? new boolean[puzzle.lang.letterCount()] : null);

		if (bestSearches.size() == 0) {
		    newSearch = new GeneticSearch(problem, stopFlag);		
		} else {

		    int r = problem.rand.nextInt(3);
		    
		    switch (r) 
		    {
		    case 0: {
			GeneticSearch parent = bestSearches.get(problem.rand.nextInt(bestSearches.size()));
			newSearch = makeMutatedSearch(parent);
			break;
		    }
		    case 1: {
			GeneticSearch parent = bestSearches.get(problem.rand.nextInt(bestSearches.size()));
			newSearch = parent.mutate(problem.rand.nextInt(puzzle.lang.letterCount()/2));
			break;
		    }
		    case 2: {
			newSearch = new GeneticSearch(problem, stopFlag);
			break;
		    }
		    }
		}

		if (newSearch == null)
		    continue;

		newSearch.listeners = listeners;
		newSearch.iterate(10000, protect);

		problem.scorethresh = Math.max(problem.scorethresh, newSearch.score);

		if (bestSearches.size()==0) {
		    bestSearches.add(newSearch);
		} else {
		    bestSearches.add(newSearch);

		    // prune searches by removing those that are insufficiently diverse
		    for (int i = 0; i < bestSearches.size(); i++) {
			for (int j = 0; j < i; j++) {
			    GeneticSearch gsi = bestSearches.get(i);
			    GeneticSearch gsj = bestSearches.get(j);

			    if (gsi == null || gsj == null)
				continue;

			    if (similarity(gsi.map, gsj.map) > 0.65) {
				if (gsi.score > gsj.score)
				    bestSearches.set(j, null);
				else
				    bestSearches.set(i, null);
			    }
			}
		    }

		    // filter out any of the NULLs
		    ArrayList<GeneticSearch> filteredSearches = new ArrayList<GeneticSearch>();
		    for (GeneticSearch gs : bestSearches)
			if (gs != null)
			    filteredSearches.add(gs);
		    bestSearches = filteredSearches;

		    // drop the worst search if necessary
		    if (bestSearches.size() > 15) {
			GeneticSearch worstSearch = null;
			int worstidx = -1;
			for (int i = 0; i < bestSearches.size(); i++) {
			    GeneticSearch gs = bestSearches.get(i);
			    
			    if (gs != null && (worstSearch == null || gs.score < worstSearch.score)) {
				worstidx = i;
				worstSearch = gs;
			    }
			}

			bestSearches.remove(worstidx);
		    }

		    /*
		    if (problem.rand.nextInt(1000)==0) {
			System.out.println("*******************************************************");
			for (GeneticSearch gs : bestSearches) {
			    System.out.println(puzzle.lang.translateString(puzzle.cipherText, new MapSet(gs.map), null));
			}
		    }
		    */
		}
	    }
	    
	    endTime = System.currentTimeMillis()/1000.0;
	    double elapsedTime = endTime - startTime;
	    
	    for (SolverListener sl : listeners)
		sl.solverFinished(elapsedTime);
	}
    }

    double similarity(int mapa[], int mapb[])
    {
	int same = 0;

	if (mapa==null || mapb==null)
	    return 0;

	for (int i = 0; i < puzzle.lang.letterCount(); i++) {
	    if (puzzle.histogram[i]>0 && mapa[i]==mapb[i])
		same++;
	}
	return ((double) same)/puzzle.numUniqueCiphertextLetters;
    }

}
